# Copyright 2021 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common selectors for evolutionary algorithms."""

import math
import random
from typing import Any, List, Union

import pyglove.core as pg
from pyglove.ext import scalars
from pyglove.ext.evolution import base


# We disable implicit str concat as it is commonly used class schema docstr.
# pylint: disable=implicit-str-concat


def num_output_spec():
  """Returns the value spec for the number of elements in the output."""
  return scalars.scalar_spec(pg.typing.Union([
      pg.typing.Int(min_value=0),                      # Output count.
      pg.typing.Float(min_value=0.0, max_value=1.0)    # Output proportion.
  ])).noneable()


def compute_num_output(
    n: Union[int, float, None], num_inputs: int, step: int) -> int:
  """Returns the number of output."""
  n = scalars.scalar_value(n, step)
  if isinstance(n, float):
    n = math.ceil(n * num_inputs)
  elif n is None:
    n = num_inputs
  return n


@pg.members([
    ('n', num_output_spec(),
     'Number of items to select, which can be an integer, a float value or '
     'None. It also can be a callable object that returns a value based on a '
     'step (an integer). If its value or returned value is an integer, the '
     'selector outputs `n` examples from the input. If float, which must be '
     'within range [0., 1.], the selector outputs `math.ceil(n * len(inputs))` '
     'examples. If None, the selector outputs the same number of examples as '
     'the input.'),
    ('replacement', pg.typing.Bool(False),
     'If True, the output examples are generated randomly with replacement. '
     'Otherwise they will be generated randomly without replacement.'),
    ('seed', pg.typing.Int().noneable(), 'Random seed for sampling the input.')
])
class Random(base.Selector):
  """Random N selector."""

  def _on_bound(self):
    super()._on_bound()
    self._random = random if self.seed is None else random.Random(self.seed)

  def select(self, inputs: List[Any], step: int) -> List[Any]:
    n = compute_num_output(self.n, len(inputs), step)
    if self.replacement:
      return [self._random.choice(inputs) for _ in range(n)]
    else:
      n = min(n, len(inputs))
      return self._random.sample(inputs, n)


@pg.members([
    ('n', num_output_spec(),
     'Number of items to select, which can be an integer, a float value or '
     'None. It also can be a callable object that returns a value based on a '
     'step (an integer). If its value or returned value is an integer, the '
     'selector outputs `n` examples from the input. If float, which must be '
     'within range [0., 1.], the selector outputs `math.ceil(n * len(inputs))` '
     'examples. If None, the selector outputs the same number of examples as '
     'the input.'),
    ('weights', base.operation_spec(
        pg.typing.Any(), pg.typing.Float(min_value=0.0)),
     'A callable object that takes a list of items as input and returns a list '
     'of float numbers as the weights for each item. Optional keyword '
     'arguments \'global_state\' and \'step\' can be accepted.'),
    ('seed', pg.typing.Int().noneable(), 'Random seed for sampling the input.')
])
class Sample(base.Selector):
  """Sample N items from a weighting function."""

  def _on_bound(self):
    super()._on_bound()
    self._random = random if self.seed is None else random.Random(self.seed)
    self._weights = base.make_operation_compatible(self.weights)

  def select(  # pytype: disable=signature-mismatch  # overriding-parameter-type-checks
      self, inputs: List[Any], global_state: pg.Dict, step: int) -> List[Any]:
    n = compute_num_output(self.n, len(inputs), step)
    weights = self._weights(inputs, global_state=global_state, step=step)
    return self._random.choices(inputs, weights=weights, k=n)


@pg.members([
    ('n', num_output_spec(),
     'Number of items to select, which can be an integer, a float value or '
     'None. It also can be a callable object that returns a value based on a '
     'step (an integer). If its value or returned value is an integer, the '
     'selector outputs `n` examples from the input. If float, which must be '
     'within range [0., 1.], the selector outputs `math.ceil(n * len(inputs))` '
     'examples. If None, the selector outputs the same number of examples as '
     'the input.'),
    ('weights', base.operation_spec(
        pg.typing.Any(), pg.typing.Float(min_value=0.0)),
     'A callable object that takes a list of items as input and returns a list '
     'of float numbers as the weights for each item. Optional keyword '
     'arguments \'global_state\' and \'step\' can be accepted.'),
])
class Proportional(base.Selector):
  """Select N items proportional to the input weights."""

  def _on_bound(self):
    super()._on_bound()
    self._weights = base.make_operation_compatible(self.weights)

  def select(  # pytype: disable=signature-mismatch  # overriding-parameter-type-checks
      self, inputs: List[Any], global_state: pg.Dict, step: int) -> List[Any]:
    n = compute_num_output(self.n, len(inputs), step)
    weights = self._weights(inputs, global_state=global_state, step=step)
    partition_info = self._partition(weights, n)
    output = []
    for item, num_occurences in zip(inputs, partition_info):
      output.extend([item] * num_occurences)
    return output

  def _partition(self, weights: List[float], n: int) -> List[int]:
    """Partition N into a list of integers roughly proportional to `weights`."""
    denominator = sum(weights)
    allocation = [int(n * w / denominator + 0.5) for w in weights]
    extra = n - sum(allocation)
    if extra != 0:
      # NOTE(daiyip): The allocation is computed by rounding up the weights
      # multiplying N, which can be less or larger than N. Therefore, we need
      # to make adjustment when such mismatch happends.
      # The strategy here is: always allocate extra items (when `extra` > 0) to
      # the slots that have larger weights, or reduce allocations
      # (when `extra` < 0) from the slots that have smaller weights.
      # To do so, we sort all the slots whose weights are not zero, and adjust
      # their allocation in a round-robin fashion.
      candidates_to_adjust = sorted(
          [i for i, w in enumerate(weights) if w > 0.],
          key=lambda x: weights[x], reverse=extra > 0)

      delta = 1 if extra > 0 else -1
      next_candidate = 0
      while extra != 0:
        item_index = candidates_to_adjust[
            next_candidate % len(candidates_to_adjust)]
        while delta == -1 and allocation[item_index] == 0:
          next_candidate = (next_candidate + 1) % len(candidates_to_adjust)
          item_index = candidates_to_adjust[next_candidate]
        allocation[item_index] += delta
        extra -= delta
        next_candidate += 1
    return allocation


@pg.members([
    ('n', num_output_spec(),
     'Number of items to select, which can be an integer, a float value or '
     'None. It also can be a callable object that returns a value based on a '
     'step (an integer). If its value or returned value is an integer, the '
     'selector outputs `n` examples from the input. If float, which must be '
     'within range [0., 1.], the selector outputs `math.ceil(n * len(inputs))` '
     'examples. If None, the selector outputs the same number of examples as '
     'the input.'),
    ('key', pg.typing.Callable(
        [pg.typing.Any()], returns=pg.typing.Any()).noneable(),
     'A callable object as the key argument for sorting the input list. '
     'If None and when the input element type is DNA, the fitness will be '
     'used as `key`.'),
    ('cluster', pg.typing.Bool(default=False),
     'If True, returns bottom N clusters. Otherwise returns N individuals. '
     'Individuals that produces the same key form a cluster.')
])
class Top(base.Selector):
  """Top N selector."""

  def select(self, inputs: List[Any], step: int) -> List[Any]:
    key = self.key
    if key is None and inputs and isinstance(inputs[0], pg.DNA):
      key = base.get_fitness
    n = compute_num_output(self.n, len(inputs), step)
    if self.cluster:
      keys = [key(x) for x in inputs]
      sorted_keys = set(sorted(set(keys), reverse=True)[:n])
      selected = [(k, x) for k, x in zip(keys, inputs)
                  if k in sorted_keys]
      return [x for k, x in sorted(
          selected, key=lambda x: x[0], reverse=True)]
    else:
      return sorted(inputs, key=key, reverse=True)[:n]


@pg.members([
    ('n', num_output_spec(),
     'Number of items to select, which can be an integer, a float value or '
     'None. It also can be a callable object that returns a value based on a '
     'step (an integer). If its value or returned value is an integer, the '
     'selector outputs `n` examples from the input. If float, which must be '
     'within range [0., 1.], the selector outputs `math.ceil(n * len(inputs))` '
     'examples. If None, the selector outputs the same number of examples as '
     'the input.'),
    ('key', pg.typing.Callable(
        [pg.typing.Any()], returns=pg.typing.Any()).noneable(),
     'A callable object as the key argument for sorting the input list. '
     'If None and when the input element type is DNA, the fitness will be '
     'used as `key`.'),
    ('cluster', pg.typing.Bool(default=False),
     'If True, returns bottom N clusters. Otherwise returns N individuals. '
     'Individuals that produces the same key form a cluster.')
])
class Bottom(base.Selector):
  """Bottom N selector."""

  def select(self, inputs: List[Any], step: int) -> List[Any]:
    key = self.key
    if key is None and inputs and isinstance(inputs[0], pg.DNA):
      key = base.get_fitness
    n = compute_num_output(self.n, len(inputs), step)
    if self.cluster:
      keys = [key(x) for x in inputs]
      sorted_keys = set(sorted(set(keys))[:n])
      selected = [(k, x) for k, x in zip(keys, inputs)
                  if k in sorted_keys]
      return [x for k, x in sorted(selected, key=lambda x: x[0])]
    else:
      return sorted(inputs, key=key)[:n]


@pg.members([
    ('n', num_output_spec(),
     'Number of items to select, which can be an integer, a float value or '
     'None. It also can be a callable object that returns a value based on a '
     'step (an integer). If its value or returned value is an integer, the '
     'selector outputs `n` examples from the input. If float, which must be '
     'within range [0., 1.], the selector outputs `math.ceil(n * len(inputs))` '
     'examples. If None, the selector outputs the same number of examples as '
     'the input.'),
])
class First(base.Selector):
  """First N selector."""

  def select(self, inputs: List[Any], step: int) -> List[Any]:
    return inputs[:compute_num_output(self.n, len(inputs), step)]


@pg.members([
    ('n', num_output_spec(),
     'Number of items to select, which can be an integer, a float value or '
     'None. It also can be a callable object that returns a value based on a '
     'step (an integer). If its value or returned value is an integer, the '
     'selector outputs `n` examples from the input. If float, which must be '
     'within range [0., 1.], the selector outputs `math.ceil(n * len(inputs))` '
     'examples. If None, the selector outputs the same number of examples as '
     'the input.')
])
class Last(base.Selector):
  """Last N selector."""

  def select(self, inputs: List[Any], step: int) -> List[Any]:
    n = compute_num_output(self.n, len(inputs), step)
    return inputs[max(0, len(inputs) - n):]



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Patcher: modular rule-based patching."""

import re
import typing
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from pyglove.core import object_utils
from pyglove.core import symbolic
from pyglove.core import typing as pg_typing


class Patcher(symbolic.Functor):
  """Class that patches a symbolic value by returning a rebind dict.

  To accomondate reusable patching, we introduced the concept of patcher,
  which is a symbolic function that takes an input value with a list of
  optional arguments and produces a patching rule (see :func:`pyglove.patch`)
  or a tuple of (patching_rule, validation_rule). Validation rule is a callable
  object that validate the patched object's integrity if it's being patched
  by others later. A patcher can be created from a URI-like string, to
  better serve the command-line interface.

  A patcher can be created via ``pg.patcher`` decorator::

    @pg.patcher([
        ('lr', pg.typing.Float(min_value=0.0))
    ])
    def learning_rate(trainer, lr):
      return {
          'training.learning_rate': lr
      }

  OR:

    @pg.patcher([
        ('lr', pg.typing.Float(min_value=0.0))
    ])
    def learning_rate(trainer, lr):
      def _rebind_fn(k, v, p):
         if k and k.key == 'learning_rate' and isinstance(v, float):
           return lr
         return v
      return _rebind_fn

  OR a composition of rules.

    @pg.patcher([
        ('lr', pg.typing.Float(min_value=0.0)),
        ('weight_decay', pg.typing.Float(min_value=0.0))
    ])
    def complex(trainer, lr, weight_decay):
      # `change_lr` and `change_weight_decay` are instances of other patchers.
      return [
          change_lr(lr),
          change_weight_decay(weight_decay)
      ]

  After registration, a patcher object can be obtained from a URL-like string::

    patcher = pg.patching.from_uri('cosine_decay?lr=0.1')

  Then the user can use the patcher to patch an object::

    patched_trainer = patcher.patch(trainer)

  A list of patchers can be applied sequentially to accomondate combination of
  semantic groups. A patcher in the sequence can propose updates to the original
  value or generate a replacement to the original value. ``pg.patching.patch``
  is introduced for making it convenient to chain multiple patchers using
  URI-like strings::

    pg.patching.patch(trainer, [
        'cosine_decay?lr=0.1',
        'some_other_patcher_string',
    ])

  The user can lookup all registered patchers via::

    print(pg.patching.patcher_names)
  """

  def _on_bound(self):
    super()._on_bound()
    self._validator = None

  def patch(self, x: symbolic.Symbolic) -> Any:
    """Patches an input and return the input itself unless fully replaced."""
    # Placeholder for Google-internal usage instrumentation.

    if not isinstance(x, symbolic.Symbolic):
      raise TypeError(
          f'The 1st argument of {self.__class__.__name__!r} must be a '
          f'symbolic type. Encountered: {x!r}.')
    # The patching rule returned from the patcher body should be either
    # (patching_rule, validation_rule) or just patching_rule.
    # The patching rule can be a dict, a rebind function, another patcher
    # or a mixture of them as a list. The patching rule is then passed to
    # ``pg.patch`` which allows patchers to support composition of sub-patchers.
    patching_rule = self.__call__(x)

    # Set validator if applicable.
    validator = None
    if isinstance(patching_rule, tuple) and len(patching_rule) == 2:
      patching_rule, validator = patching_rule
    if validator is not None and not callable(validator):
      raise TypeError(
          f'The validator returned from patcher {self.__class__.__name__!r} '
          f'is not callable. Encountered: {validator!r}.')
    self._validator = validator
    return patch(x, patching_rule)

  def validate(self, x: symbolic.Symbolic) -> None:
    """Validates an input's integrity.

    This method will be called in :func:`pyglove.patch` when a chain of patchers
    have been applied, as to validate the patched object in chain still
    conforms to the patcher's plan.

    Args:
      x: The input after modification.
    """
    if self._validator is not None:
      self._validator(x)

  def __call__(
      self,
      x: symbolic.Symbolic
      ) -> Union[Dict[str, Any], Tuple[Dict[str, Any], Callable[[], None]]]:
    """Override __call__ to get rebind dict."""
    return super().__call__(x, override_args=True)


_ALLOW_REPEATED_PATCHER_REGISTRATION = True


def allow_repeated_patcher_registration(allow: bool = True):
  """If True, allow registration with the same patch name."""
  global _ALLOW_REPEATED_PATCHER_REGISTRATION
  _ALLOW_REPEATED_PATCHER_REGISTRATION = allow


class _PatcherRegistry:
  """Patcher registry."""

  def __init__(self):
    self._registry = dict()

  def get(self, name) -> Type[Patcher]:
    """Returns patch class by name."""
    if name not in self._registry:
      raise KeyError(f'Patcher {name!r} is not registered.')
    return self._registry[name]

  def register(self, name: str, patcher_cls: Type[Patcher]):
    """Register a function with a scheme name."""
    if name in self._registry and not _ALLOW_REPEATED_PATCHER_REGISTRATION:
      raise KeyError(f'Patcher {name!r} already registered.')
    self._registry[name] = patcher_cls

  @property
  def names(self) -> List[str]:
    """Returns registered scheme names."""
    return list(self._registry.keys())


_PATCHER_REGISTRY = _PatcherRegistry()


def patcher(
    args: Optional[List[Tuple[str, pg_typing.ValueSpec]]] = None,
    name: Optional[str] = None) -> Any:
  """Decorate a function into a Patcher and register it.

  A patcher function is defined as:

     `<patcher_fun> := <fun_name>(<target>, [parameters])`

  The signature takes at least one argument as the patching target,
  with additional arguments as patching parameters to control the details of
  this patch.

  Example::

    @pg.patching.patcher([
      ('x': pg.typing.Int())
    ])
    def increment(v, x=1):
      return pg.symbolic.get_rebind_dict(
          lambda k, v, p: v + x if isinstance(v, int) else v)

    # This patcher can be called via:
    # pg.patching.apply(v, [increment(x=2)])
    # or pg.patching.apply(v, ['increment?x=2'])

  Args:
    args: A list of (arg_name, arg_value_spec) to schematize patcher arguments.
    name: String to be used as patcher name in URI. If None, function name will
      be used as patcher name.

  Returns:
    A decorator that converts a function into a Patcher subclass.
  """
  functor_decorator = symbolic.functor(args, base_class=Patcher)
  def _decorator(fn):
    """Returns decorated Patcher class."""
    cls = functor_decorator(fn)
    _PATCHER_REGISTRY.register(name or fn.__name__,
                               typing.cast(Type[Patcher], cls))
    arg_specs = cls.signature.args
    if len(arg_specs) < 1:
      raise TypeError(
          f'Patcher function should have at least 1 argument '
          f'as patching target. (Patcher={cls.type_name!r})')
    if not _is_patcher_target_spec(arg_specs[0].value_spec):
      raise TypeError(
          f'{arg_specs[0].value_spec!r} cannot be used for constraining '
          f'Patcher target. (Patcher={cls.type_name!r}, '
          f'Argument={arg_specs[0].name!r})\n'
          f'Acceptable value spec types are: '
          f'Any, Callable, Dict, Functor, List, Object.')
    for arg_spec in arg_specs[1:]:
      if not _is_patcher_parameter_spec(arg_spec.value_spec):
        raise TypeError(
            f'{arg_spec.value_spec!r} cannot be used for constraining '
            f'Patcher argument. (Patcher={cls.type_name!r}, '
            f'Argument={arg_spec.name!r})\n'
            f'Consider to treat it as string and parse yourself.')
    return cls
  return _decorator


def _is_patcher_target_spec(value_spec):
  """Return True if value_spec can be used for patcher target."""
  return isinstance(
      value_spec, (pg_typing.Any, pg_typing.Object,
                   pg_typing.Dict, pg_typing.List, pg_typing.Callable))


def _is_patcher_parameter_spec(value_spec, leaf_only=False):
  """Return True if value_spec can be used for patcher parameters."""
  if isinstance(value_spec, (pg_typing.Any, pg_typing.Str, pg_typing.Bool,
                             pg_typing.Int, pg_typing.Float)):
    return True
  elif isinstance(value_spec, pg_typing.Enum):
    return value_spec.value_type == str
  elif isinstance(value_spec, pg_typing.List):
    return (not leaf_only  and _is_patcher_parameter_spec(
        value_spec.element.value, leaf_only=True))
  return False


def patcher_names():
  """Returns all registered patch names."""
  return _PATCHER_REGISTRY.names


PatchType = Union[
    Dict[str, Any], Callable, Patcher, str,               # pylint: disable = g-bare-generic
    List[Union[Dict[str, Any], Callable, Patcher, str]]]  # pylint: disable = g-bare-generic


def patch(value: symbolic.Symbolic, rule: PatchType) -> Any:
  """Apply patches to a symbolic value.

  Args:
    value: A symbolic value to patch.
    rule: A patching rule is one of the following:
      1) A dict of symbolic paths to the new values.
      2) A rebind function defined by signature (k, v) or (k, v, p).
         See :meth:`pyglove.Symbolic.rebind`.
      3) A :class:`pyglove.patching.Patcher` object.
      4) A URL-like string representing an instance of a register Patcher.
         The format is "<patcher_name>?<arg1>&<arg2>=<val2>".
      5) A list of the mixtures of above.

  Returns:
    Value after applying the patchers. If any patcher returned a new value
    (by returning a single-item dict that containing '' as key), the return
    value will be a different object other than `value`, otherwise `value`
    will be returned after applying the patches.

  Raises:
    ValueError: Error if the patch name and arguments cannot
      be parsed successfully.
  """
  patches = []
  rules = rule if isinstance(rule, list) else [rule]
  for p in rules:
    if isinstance(p, str):
      p = from_uri(p)
    if not isinstance(p, (Patcher, dict)) and not callable(p):
      raise TypeError(
          f'Patching rule {p!r} should be a dict of path to values, a rebind '
          f'function, a patcher (object or string), or a list of their '
          f'mixtures.')
    patches.append(p)

  # Apply patches in chain.
  for p in patches:
    if isinstance(p, Patcher):
      value = p.patch(value)
    elif isinstance(p, dict):
      if len(p) == 1 and '' in p:
        value = p['']
      else:
        value = value.rebind(p, raise_on_no_change=False)
    else:
      value = value.rebind(p, raise_on_no_change=False)

  # Validate patched values.
  for p in patches:
    if isinstance(p, Patcher):
      p.validate(value)
  return value


_ID_REGEX = re.compile(r'^[a-zA-Z_][a-zA-Z0-9_]*')
_ARG_ASSIGN_REGEX = re.compile(r'^([a-zA-Z_][a-zA-Z0-9_]*)=(.*)')


def from_uri(uri: str) -> Patcher:
  """Create a Patcher object from a URI-like string."""
  name, args, kwargs = parse_uri(uri)
  patcher_cls = typing.cast(Type[Any], _PATCHER_REGISTRY.get(name))
  args, kwargs = parse_args(patcher_cls.signature, args, kwargs)
  return patcher_cls(object_utils.MISSING_VALUE, *args, **kwargs)


def parse_uri(uri: str) -> Tuple[str, List[str], Dict[str, str]]:
  """Parse patcher name and arguments from a URI-like string."""
  pos = uri.find('?')
  args = []
  kwargs = dict()
  if pos >= 0:
    name = uri[:pos]
    args_str = uri[pos + 1:]
    for arg_str in args_str.split('&'):
      if '=' in arg_str:
        m = _ARG_ASSIGN_REGEX.match(arg_str)
        if m is None:
          raise ValueError(f'Invalid argument specification: {arg_str}.')
        assert len(m.groups()) == 2
        arg_name, arg_value = m.groups()
        assert _ID_REGEX.match(arg_name)
        kwargs[arg_name] = arg_value
      else:
        if kwargs:
          raise ValueError(
              f'Positional argument should be provided before keyword '
              f'arguments. Encountered: {uri!r}.')
        args.append(arg_str)
  else:
    name = uri

  if not _ID_REGEX.match(name):
    raise ValueError(f'{name!r} is not a valid Patcher name.')
  return name, args, kwargs    # pytype: disable=bad-return-type


def parse_args(signature: pg_typing.Signature,
               args: List[str],
               kwargs: Dict[str, str]) -> Tuple[List[Any], Dict[str, Any]]:
  """Parse patcher arguments based on its signature."""
  acceptable_arg_names = [arg.name for arg in signature.args[1:]]
  if len(signature.args) < len(args) + 1:
    raise KeyError(
        f'Too many positional arguments are provided. '
        f'(Patcher={signature.id})\n'
        f'Expected: {acceptable_arg_names!r}, Received: {args!r}.')
  kw = {arg.name: arg for arg in signature.args}
  parsed_args = []
  parsed_kwargs = {}
  for i, arg_str in enumerate(args):
    arg_spec = signature.args[i + 1]
    parsed_args.append(
        parse_arg(signature.id, arg_spec.name, arg_spec.value_spec, arg_str))
  for k, arg_str in kwargs.items():
    if k not in kw:
      raise KeyError(
          f'Unexpected argument {k!r}. (Patcher={signature.id!r})\n'
          f'Acceptable argument names: {acceptable_arg_names!r}.')
    arg_spec = kw[k]
    parsed_kwargs[k] = parse_arg(
        signature.id, arg_spec.name, arg_spec.value_spec, arg_str)
  return parsed_args, parsed_kwargs


_BOOL_TRUE_LITERAL_SET = frozenset(['True', 'true', 'yes', '1'])
_BOOL_FALSE_LITERAL_SET = frozenset(['False', 'false', 'no', '0'])
_NONE_LITERAL_SET = frozenset(['None', 'none'])


def parse_arg(patcher_id: str, arg_name: str,
              value_spec: pg_typing.ValueSpec, arg_str: str):
  """Parse patcher argument based on value spec."""
  def _value_error(msg):
    return ValueError(f'{msg} (Patcher={patcher_id!r}, Argument={arg_name!r})')

  if arg_str in _NONE_LITERAL_SET:
    # NOTE(daiyip): If string type value needs literal 'None' or 'none', they
    # can quote the string with "".
    arg = None
  elif isinstance(value_spec, (pg_typing.Any, pg_typing.Str)):
    if len(arg_str) > 1:
      begin_quote = (arg_str[0] == '"')
      end_quote = (arg_str[-1] == '"')
      if begin_quote and end_quote:
        arg_str = arg_str[1:-1]
      elif begin_quote or end_quote:
        raise _value_error(f'Unmatched quote for string value: {arg_str!r}.')
    arg = arg_str
  elif isinstance(value_spec, pg_typing.Bool):
    if (arg_str not in _BOOL_TRUE_LITERAL_SET
        and arg_str not in _BOOL_FALSE_LITERAL_SET):
      raise _value_error(f'Cannot convert {arg_str!r} to bool.')
    arg = arg_str in _BOOL_TRUE_LITERAL_SET
  elif isinstance(value_spec, pg_typing.Int):
    try:
      arg = int(arg_str)
    except ValueError:
      raise _value_error(f'Cannot convert {arg_str!r} to int.')  # pylint: disable=raise-missing-from
  elif isinstance(value_spec, pg_typing.Float):
    try:
      arg = float(arg_str)
    except ValueError:
      raise _value_error(f'Cannot convert {arg_str!r} to float.')  # pylint: disable=raise-missing-from
  elif isinstance(value_spec, pg_typing.Enum):
    if value_spec.value_type != str:
      raise _value_error(
          f'{value_spec!r} cannot be used for Patcher argument. '
          f'Only Enum of string type can be used.')
    arg = arg_str
  elif isinstance(value_spec, pg_typing.List):
    arg = parse_list(
        arg_str,
        lambda i, s: parse_arg(    # pylint: disable=g-long-lambda
            patcher_id, f'{arg_name}[{i}]', value_spec.element.value, s))
  else:
    raise _value_error(
        f'{value_spec!r} cannot be used for Patcher argument.\n'
        f'Consider to treat this argument as string and parse it yourself.')
  return value_spec.apply(
      arg, root_path=object_utils.KeyPath.parse(f'{patcher_id}.{arg_name}'))


def parse_list(string: str,
               convert_fn: Callable[[int, str], Any]) -> List[Any]:
  """Parse a (possibly empty) colon-separated list of values."""
  string = string.strip()
  if string:
    return [convert_fn(i, piece) for i, piece in enumerate(string.split(':'))]
  return []




# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Sub-module for exposing PyGlove DNA generators."""

# NOTE(daiyip): We disable bad-import-order to preserve the relation of
# imported symbols
# pylint: disable=g-import-not-at-top
# pylint: disable=g-bad-import-order

from pyglove import core
from pyglove import ext

# NOTE(daiyip): For backward compatibility.
evolution = ext.evolution
RegularizedEvolution = ext.evolution.RegularizedEvolution
evolution_mutators = ext.evolution.mutators

geno = core.geno
Sweeping = geno.Sweeping
Random = geno.Random
Deduping = geno.Deduping

# pylint: enable=g-import-not-at-top
# pylint: enable=g-bad-import-order



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Distributed sampling."""

from typing import Any, Callable, Optional, Sequence, Union

from pyglove.core import geno
from pyglove.core import hyper
from pyglove.core import symbolic
from pyglove.core.tuning import backend as backend_lib
from pyglove.core.tuning.early_stopping import EarlyStoppingPolicy


def sample(hyper_value: Any,
           algorithm: geno.DNAGenerator,
           num_examples: Optional[int] = None,
           early_stopping_policy: Optional[EarlyStoppingPolicy] = None,
           where: Optional[Callable[[hyper.HyperPrimitive], bool]] = None,
           name: Optional[str] = None,
           group: Union[None, int, str] = None,
           backend: Optional[str] = None,
           metrics_to_optimize: Optional[Sequence[str]] = None,
           **kwargs):
  """Yields an example and its feedback sampled from a hyper value.

  Example 1: sample a search space defined by a symbolic hyper value::

    for example, feedback in pg.sample(
        hyper_value=pg.Dict(x=pg.floatv(-1, 1))),
        algorithm=pg.generators.Random(),
        num_examples=10,
        name='my_search'):

      # We can access trial ID (staring from 1) and DNA from the feedback.
      print(feedback.id, feedback.dna)

      # We can report the reward computed on the example using
      # `feedback.add_measurement`, which can be called
      # multiple times to report the rewards incrementally.
      # Once a trial is done, we call `feedback.done` to mark evaluation on
      # current example as completed, or use `feedback.skip` to move to the
      # next sample without passing any feedback to the algorithm.
      # Without `feedback.done` or `feedback.skip`, the same trial will be
      # iterated over and over again for failover handling purpose.
      # Besides the reward and the step, metrics and checkpoint can be added
      # to each measurement. Additional meta-data and related links (URLs) can
      # be passed to `feedback.done` which can be retrieved via
      # `pg.poll_result` later.
      if example.x >= 0:
        # If we only want to add one measurement for each example, a
        # shortcut expression for the next two lines can be written as
        # follows:
        #   `feedback(reward=math.sqrt(example.x), step=1)`
        feedback.add_measurement(reward=math.sqrt(example.x), step=1)
        feedback.done()
      else:
        feedback.skip()

      # IMPORTANT: to stop the loop among all workers, we can call
      # `feedback.end_loop`. As a result, each worker will quit their loop
      # after current iteration, while using `break` in the for-loop is
      # only effective for the local process rather than remotely.
      if session.id > 1000:
        feedback.end_loop()

    # At any time, we can poll the search result via `pg.poll_result`, please
    # see `pg.tuning.Result` for more details.
    result = pg.poll_result('my_search')
    print(result.best_trial)

  Example 2: sample a search space defined by `pg.hyper.trace`::

    def fun():
      return pg.oneof([
        lambda: pg.oneof([1, 2, 3]),
        lambda: pg.float(0.1, 1.0),
        3]) + sum(pg.manyof(2, [1, 2, 3]))

    for example, feedback in pg.sample(
        hyper_value=pg.hyper.trace(fun),
        algorithm=pg.generators.Random(),
        num_examples=10,
        name='my_search'):
      # When hyper_value is a `pg.hyper.DynamicEvaluationContext` object,
      # the `example` yielded at each iteration is a context manager under which
      # the hyper primitives (e.g. pg.oneof) will be materialized into concrete
      # values according to the controller decision.
      with example():
        reward = fun()
      feedback(reward)

  **Using `pg.sample` in distributed environment**

  `pg.sample` is designed with distributed sampling in mind, in which multiple
  processes can work on the trials of the same sampling loop. While the default
  'in-memory' backend works only within a single process without failover
  handling, other backends may support distributed computing environment with
  persistent state. Nevertheless, the `pg.sample` API is the same
  among different backends, users can switch the backend easily by passing a
  different value to the `backend` argument, or set the default value globally
  via `pg.tuning.set_default_backend`.


  **Identifying a sampling**

  To identify a distributed loop, a unique `name` is introduced, which will
  also be used to poll the latest sampling result via `pg.poll_result`.

  **Failover handling**

  In a distributed setup, worker processes may incidentally die and restart.
  Unless a trial is explicitly marked as done (via `feedback(reward)` or
  `feedback.done()`) or skipped (via feedback.skip()), a worker will try to
  resume its work on the trial from where it left off.

  **Workroup**

  In a distributed setup, worker processes may or may not work on the same
  trials. Worker group is introduced to serve this purpose, which is identified
  by an integer or a string named `group`. If `group` is not specified or
  having different values among the workers, every worker will work on
  different trials of the loop. On the contrary, workers having the same
  `group` will co-work on the same trials. Group is useful when evaluation on
  one example can be parallelized - for example - an example in the outer loop
  of a nested search. However, feedback on the same example should be fed back
  to the search algorithm only once. Therefore, workers in the same group need
  to communicate with each other to avoid duplicated evaluation and feedbacks.
  To faciliate such communication, per-trial metadata is supported, and can be
  accessed via `feedback.set_metadata/get_metadata` methods. The consistency in
  reading and writing the metadata is defined by the backend used. For the
  'in-memory' backend, all the trials and their metadata is stored in memory,
  thus will be lost if the process get restarted. On the contrary, the backends
  built on distributed computing environment may store both the trials and
  metadata in a persistent storage with varing read/write QPS and read/write
  consistency guarentees.

  **Switch between backends**

  The `backend` argument of `pg.sample` lets users choose a backend used in
  current sampling loop. Users can use different backends in the same process
  to achieve a best performance trade-off, e.g., using `in-memory` backend when
  the communication cost overweighs the redudant evaluation cost upon worker
  failure.

  Helper function :func:`pyglove.tuning.set_default_backend` is introduced to
  set the default tuning backend for the entire process.

  Args:
    hyper_value: A hyper value to sample from. A hyper value is an object with
      to-be-determined values specified by `pg.oneof`, `pg.manyof`, `pg.floatv`
      and etc, representing a search space, or a
      `pg.hyper.DynamicEvaluationContext` object.
    algorithm: The search algorithm that samples the search space. For example:
      `pg.generators.Random()`, `pg.evolution.regularized_evolution(...)`, and
      etc.
    num_examples: An optional integer as the max number of examples to
      sample. If None, sample will return an iterator of infinite examples.
    early_stopping_policy: An optional early stopping policy for user to tell
      if incremental evaluation (which reports multiple measurements) on each
      example can be early short circuited.
      After each call to `feedback.add_measurement`, users can use method
      `feedback.should_stop_early` to check whether current example worth
      further evaluation or not.
    where: Function to filter the hyper values. If None, all decision points
      from the `hyper_value` will be included for the algorithm to make
      decisions. Otherwise only the decision points on which 'where' returns
      True will be included. The rest decision points will be passed through
      in the example, intact, which is a sub-space of the search space
      represented by the `hyper_value`. `where` is usually used in nested search
      flows. Please see 'hyper.Template' docstr for details.
    name: A string as a unique identifier for current sampling. Two separate
      calls to `pg.sample` with the same `name` (also the same algorithm) will
      share the same sampling queue, whose examples are proposed by the same
      search algorithm.
    group: An string or integer as the group ID of current process in
      distributed sampling, which will be used to group different workers into
      co-worker groups. Workers with the same group id will work on the same
      trial. On the contrary, workers in different groups will always be working
      with different trials. If not specified, each worker in current sampling
      will be in different groups. `group` is usually used in the outer
      loops of nested search, in order to allow workers to work on the same
      higher-order item.
    backend: An optional string to specify the backend name for sampling.
      if None, the default backend set by `pg.tuning.set_default_backend`
      will be used.
    metrics_to_optimize: A sequence of string as the names of the metrics
      to be optimized by the algorithm, which is ['reward'] by default.
      When specified, it should have only 1 item for single-objective algorithm
      and can have multiple items for algorithms that support multi-objective
      optimization.
    **kwargs: Arguments passed to the `BackendFactory` subclass registered with
      the requested backend.

  Yields:
    An iterator of tuples (example, feedback) as examples sampled from the
    search space defined by the `hyper_value` through `algorithm`.

  Raises:
    ValueError: `hyper_value` is a fixed value, or requested `backend` is not
    available.
  """
    # Placeholder for Google-internal usage instrumentation.

  # Create template based on the hyper value.
  if isinstance(hyper_value, hyper.DynamicEvaluationContext):
    dynamic_evaluation_context = hyper_value
    dna_spec = hyper_value.dna_spec
    template = None
  else:
    if symbolic.is_deterministic(hyper_value):
      raise ValueError(f'\'hyper_value\' is a constant value: {hyper_value!r}.')
    template = hyper.template(hyper_value, where)
    dna_spec = template.dna_spec()
    dynamic_evaluation_context = None

  # Create and set up the backend.
  metrics_to_optimize = metrics_to_optimize or ['reward']
  backend = backend_lib.get_backend_cls(backend).create(
      name, group, dna_spec, algorithm, metrics_to_optimize,
      early_stopping_policy, num_examples, **kwargs)

  while True:
    try:
      feedback = backend.next()
      dna = feedback.dna
      reward = dna.metadata.get('reward')
      if reward is None:
        # Decode and return current example to client code for evaluation.
        if template is not None:
          value = template.decode(dna)
        else:
          assert dynamic_evaluation_context is not None
          value = lambda: dynamic_evaluation_context.apply(dna)
        yield (value, feedback)
      else:
        # Reward may be computed at the controller side, we can
        # short circuit the client-side evaluation for the current item.
        # Also, there can be multiple co-workers working on the same trial,
        # we ignore errors triggered by racing feedbacks, which will be
        # considered as no-op.
        with feedback.ignore_race_condition():
          feedback(reward, metadata=dict(client_evaluation_skipped=True))
    except StopIteration:
      return
